!---------------------------------------------------------------------------------------
! PREP_SATELLITE: get_tropomi
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2022, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> get_tropomi
!!
!! Purpose:    Reads observations and auxilliary data from the specified satellite and
!!             and writes the output into files that can be directly used by FLEXPART
!!             and, for inversions, by FLEXINVERT
!!
!! Interface:
!!
!!    Inputs
!!             settings  -  settings data structure
!!
!!    Outputs
!!
!!    Externals
!!
!---------------------------------------------------------------------------------------

subroutine get_tropomi(settings)

  use mod_settings
  use mod_var
  use mod_dates
  use mod_save
  use netcdf

  implicit none

  type (settings_t), intent(in)                         :: settings

  character(max_name_len), dimension(:), allocatable    :: filelist, filereadlist
  character(max_name_len)                               :: line, string, grpname, filename
  character(max_name_len)                               :: suppdata_grpname, geodata_grpname, indata_grpname, details_grpname
  character(len=8)                                      :: adate
  character(len=8)                                      :: surfclass_bin
  character(len=6)                                      :: atime
  character(len=3)                                      :: aspec
  character(len=2)                                      :: amon, aday
  character(len=6), dimension(4)                        :: units
  integer                                               :: i, ierr, n, nf, nl, nm, np, nt, nbits, iswater 
  integer                                               :: nfiles, nread, nretr, numbox 
  real(kind=8)                                          :: jdatei, jdatef, jd, ndays, jref, jdmeas
  integer                                               :: jjjjmmdd, hhmiss, mm, dd, hh, mi, ss
  integer, parameter                                    :: refdate = 20100101
  integer, parameter                                    :: reftime = 0
  integer, parameter                                    :: lsynctime = 600
  integer                                               :: ntime, nmeas, npixel, nlevel, ncoord, nlayer
  real, dimension(:,:,:,:), allocatable                 :: lat_bnds, lon_bnds
  real, dimension(:,:,:), allocatable                   :: psurf, pdel, vmr, verr, qa, albedo
  real, dimension(:,:,:,:), allocatable                 :: vapri, vdryair, cak
  real, dimension(:,:), allocatable                     :: xpoints, ypoints
  real, dimension(:,:), allocatable                     :: zpoint1, zpoint2
  real, dimension(:,:), allocatable                     :: cak_out, vapri_out, vdryair_out
  real, dimension(:,:), allocatable                     :: cak_tmp, vapri_tmp, vdryair_tmp
  real, dimension(:), allocatable                       :: vmr_out, verr_out, albedo_out
  real, dimension(:), allocatable                       :: cpri, cakpri, cdryair
  integer, dimension(:), allocatable                    :: idate, itime, time
  integer, dimension(:,:), allocatable                  :: timedel
  integer, dimension(:,:,:), allocatable                :: surfclass
  integer                                               :: specnum, nchar
  integer                                               :: ncid, grpid, supp_grpid, geo_grpid, in_grpid, det_grpid
  integer                                               :: dimid, varid 
  real                                                  :: nest_llx, nest_lly, nest_urx, nest_ury
  real                                                  :: llx, lly, urx, ury
  real                                                  :: sclfact, offset 
  real, parameter                                       :: bignum=huge(0.)
  real                                                  :: const, alb, factor_corr
  real, dimension(11)                                   :: poly

  write(logid,*) 'largest number that can be handled (used to define NA) = ',bignum

  ! get species number
  ! ------------------

  open(100,file=trim(settings%path_flexpart)//'options/SPECIES/spec_overview',status='old',action='read',iostat=ierr)
  nchar = len_trim(settings%species)
  do while ( ierr.eq.0 )
    read (100, fmt='(A)', iostat=ierr) line
    if ( line(len_trim(line)-nchar+1:len_trim(line)) == trim(settings%species) ) ierr = 1
  end do
  read(line(9:11),*) specnum
  write(logid,*) 'species: ',settings%species
  write(logid,*) 'specnum: ',specnum
  close(100)

  ! define domains
  ! --------------

  llx = settings%outlonleft
  lly = settings%outlatlower
  urx = llx + settings%numxgrid*settings%dxout
  ury = lly + settings%numygrid*settings%dyout

  if ( settings%lnested.eq.1 ) then
    nest_llx = settings%outlonnest
    nest_lly = settings%outlatnest
    nest_urx = nest_llx + settings%numxnest*settings%dxoutnest
    nest_ury = nest_lly + settings%numynest*settings%dyoutnest
  endif

  ! list satellite files
  ! ----------------------
 
  call system('find '//trim(settings%path_obs)//' -type f | grep '//trim(settings%suffix) &
              //' | wc -l > '//trim(settings%path_obsout)//'obsfiles.txt',ierr)
  if ( ierr.eq.0 ) then
    call system('find '//trim(settings%path_obs)//' -type f | grep '//trim(settings%suffix) &
                //' >> '//trim(settings%path_obsout)//'obsfiles.txt')
  else
    write(logid,*) 'ERROR: cannot find observation files'
    stop
  endif

  open(100,file=trim(settings%path_obsout)//'obsfiles.txt',action='read',status='old',iostat=ierr)
  if ( ierr.ne. 0 ) then
    write(logid,*) 'ERROR: cannot open obsfiles.txt'
    stop
  endif
  read(100,*,iostat=ierr) nfiles
  allocate ( filelist(nfiles), stat = ierr )
  allocate ( filereadlist(nfiles), stat = ierr )
  do nf = 1, nfiles
    read(100,fmt='(A)',iostat=ierr) filelist(nf)
    if (ierr.ne.0) exit
  end do
  close(100)

  ! loop over days 
  ! --------------

  jdatei = juldate(settings%datei, 0)
  jdatef = juldate(settings%datef, 0)
  jd = jdatei

  read_loop: do

    if ( jd.gt.jdatef ) exit read_loop

    ! search files for current date
    call caldate(jd, jjjjmmdd, hhmiss)
    write(adate,fmt='(I8.8)') jjjjmmdd
    write(logid,*) 'current day: ',adate
    mm = jjjjmmdd/100 - (jjjjmmdd/10000)*100
    dd = jjjjmmdd - (jjjjmmdd/100)*100
    write(amon,fmt='(I2.2)') mm
    write(aday,fmt='(I2.2)') dd
    nread = 0
    do n = 1, nfiles
      ! corrupted files have "xxx" at start of name
      i=len_trim(settings%path_obs)
      if ( filelist(n)((i+1):(i+3)).eq.'xxx' ) then 
        print*, 'skipping file: ',filelist(n)
        cycle
      endif
      do i = 1, len_trim(filelist(n))-18
        if ( filelist(n)(i:(i+7)).eq.adate ) then
          nread = nread + 1
          filereadlist(nread) = filelist(n)
          exit
        endif
      end do
    end do
    write(logid,*) 'number of files for current day: ',nread
    if ( nread.eq.0 ) then
      jd = jd + 1d0
      cycle read_loop
    endif

    ! loop over files for current day
    ! --------------------------------

    do n = 1, nread

      ! open file
      write(logid,*) 'reading file: ',filereadlist(n)
      call check( nf90_open(trim(filereadlist(n)),nf90_NOWRITE,ncid) )

      ! dimension variables
      string = settings%time_name
      call split(string,'/',grpname)
      call check( nf90_inq_ncid(ncid,trim(grpname),grpid) )
      call check( nf90_inq_dimid(grpid,trim(string),dimid) )
      call check( nf90_inquire_dimension(grpid,dimid,len=ntime) )
      write(logid,*) 'number reference times: ',ntime
      string = settings%nmeas_name
      call split(string,'/',grpname)
      call check( nf90_inq_ncid(ncid,trim(grpname),grpid) )
      call check( nf90_inq_dimid(grpid,trim(string),dimid) )
      call check( nf90_inquire_dimension(grpid,dimid,len=nmeas) )
      write(logid,*) 'number of scanlines: ',nmeas
      string = settings%npixel_name
      call split(string,'/',grpname)
      call check( nf90_inq_ncid(ncid,trim(grpname),grpid) )
      call check( nf90_inq_dimid(grpid,trim(string),dimid) )
      call check( nf90_inquire_dimension(grpid,dimid,len=npixel) )
      write(logid,*) 'number of pixels: ',npixel
      string = settings%ncoord_name
      call split(string,'/',grpname)
      call check( nf90_inq_ncid(ncid,trim(grpname),grpid) )
      call check( nf90_inq_dimid(grpid,trim(string),dimid) )
      call check( nf90_inquire_dimension(grpid,dimid,len=ncoord) )
      write(logid,*) 'number of geocoordinates: ',ncoord
      string = settings%nlevel_name
      call split(string,'/',grpname)
      call check( nf90_inq_ncid(ncid,trim(grpname),grpid) )
      call check( nf90_inq_dimid(grpid,trim(string),dimid) )
      call check( nf90_inquire_dimension(grpid,dimid,len=nlevel) )
      nlayer = nlevel - 1
      write(logid,*) 'number of vertical layers: ',nlayer

      ! geolocation variables
      allocate( lat_bnds(ncoord,npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate lat_bnds'
      allocate( lon_bnds(ncoord,npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate lon_bnds'
      string = settings%lon_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',geodata_grpname)
      call check( nf90_inq_ncid(ncid,trim(grpname),grpid) )
      call check( nf90_inq_ncid(grpid,trim(suppdata_grpname),supp_grpid) )
      call check( nf90_inq_ncid(supp_grpid,trim(geodata_grpname),geo_grpid) )
      call check( nf90_inq_varid(geo_grpid,trim(string),varid) )
      call check( nf90_get_var(geo_grpid,varid,lon_bnds) )
      write(logid,*) 'read longitude'
      string = settings%lat_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',geodata_grpname)
      call check( nf90_inq_varid(geo_grpid,trim(string),varid) )
      call check( nf90_get_var(geo_grpid,varid,lat_bnds) )
      write(logid,*) 'read latitude'
      
      ! input data variables
      allocate( psurf(npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate psurf'
      allocate( pdel(npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate pdel'
      allocate( surfclass(npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate surfclass'
      allocate( vapri(nlayer,npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate vapri'
      allocate( vdryair(nlayer,npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate vdryair'
      string = settings%psurf_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',indata_grpname)
      call check( nf90_inq_ncid(ncid,trim(grpname),grpid) )
      call check( nf90_inq_ncid(grpid,trim(suppdata_grpname),supp_grpid) )
      call check( nf90_inq_ncid(supp_grpid,trim(indata_grpname),in_grpid) )
      call check( nf90_inq_varid(in_grpid,trim(string),varid) )
      call check( nf90_get_var(in_grpid,varid,psurf) )
      write(logid,*) 'read surface pressure'
      string = settings%pdel_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',indata_grpname)
      call check( nf90_inq_varid(in_grpid,trim(string),varid) )
      call check( nf90_get_var(in_grpid,varid,pdel) )
      write(logid,*) 'read pressure coordinates'
      string = settings%apri_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',indata_grpname)
      call check( nf90_inq_varid(in_grpid,trim(string),varid) )
      call check( nf90_get_var(in_grpid,varid,vapri) )
      write(logid,*) 'read prior vertical profile'
      string = settings%dryair_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',indata_grpname)
      call check( nf90_inq_varid(in_grpid,trim(string),varid) )
      call check( nf90_get_var(in_grpid,varid,vdryair) )
      write(logid,*) 'read dry air profile'
      string = settings%surfclass_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',indata_grpname)
      call check( nf90_inq_varid(in_grpid,trim(string),varid) )
      call check( nf90_get_var(in_grpid,varid,surfclass) )
      write(logid,*) 'read surface class'

      ! detailed results variables
      allocate( cak(nlayer,npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate cak'
      string = settings%ak_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',details_grpname)
      call check( nf90_inq_ncid(supp_grpid,trim(details_grpname),det_grpid) )
      call check( nf90_inq_varid(det_grpid,trim(string),varid) )
      call check( nf90_get_var(det_grpid,varid,cak) )
      write(logid,*) 'read column averaging kernel'
      allocate( albedo(npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate albedo'
      string = settings%albedo_name
      call split(string,'/',grpname)
      call split(string,'/',suppdata_grpname)
      call split(string,'/',details_grpname)
      call check( nf90_inq_ncid(supp_grpid,trim(details_grpname),det_grpid) )
      call check( nf90_inq_varid(det_grpid,trim(string),varid) )
      call check( nf90_get_var(det_grpid,varid,albedo) )
      write(logid,*) 'read surface albedo SWIR'

      ! product variables
      allocate( qa(npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate qa'
      allocate( vmr(npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate vmr'
      allocate( verr(npixel,nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate verr'
      allocate( timedel(nmeas,ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate timedel'
      allocate( time(ntime), stat=ierr )
      if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate time'
      string = settings%qa_name
      call split(string,'/',grpname)
      call check( nf90_inq_ncid(ncid,trim(grpname),grpid) )
      call check( nf90_inq_varid(grpid,trim(string),varid) )
      call check( nf90_get_var(grpid,varid,qa) )
      call check( nf90_get_att(grpid,varid,'scale_factor',sclfact) )
      call check( nf90_get_att(grpid,varid,'add_offset',offset) )
      qa = qa*sclfact + offset

      write(logid,*) 'read quality value'
      string = settings%vmr_name
      call split(string,'/',grpname)
      call check( nf90_inq_varid(grpid,trim(string),varid) )
      call check( nf90_get_var(grpid,varid,vmr) )
      write(logid,*) 'read total column mixing ratio'
      string = settings%verr_name
      call split(string,'/',grpname)
      call check( nf90_inq_varid(grpid,trim(string),varid) )
      call check( nf90_get_var(grpid,varid,verr) )
      write(logid,*) 'read total column error'
      string = settings%tdel_name
      call split(string,'/',grpname)
      call check( nf90_inq_varid(grpid,trim(string),varid) )
      call check( nf90_get_var(grpid,varid,timedel) )
      write(logid,*) 'read delta time'
      string = settings%time_name
      call split(string,'/',grpname)
      call check( nf90_inq_varid(grpid,trim(string),varid) )
      call check( nf90_get_var(grpid,varid,time) )
      write(logid,*) 'reference time = ', time

      call check( nf90_close(ncid)) 

!      print*, 'range(qa) = ',minval(qa),maxval(qa)

      ! select retrievals and extract needed info
      if( n.eq.1 ) then
        allocate( xpoints(maxretr,ncoord), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate xpoints'
        allocate( ypoints(maxretr,ncoord), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate ypoints'
        allocate( zpoint1(maxretr,nlayer), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate zpoint1'
        allocate( zpoint2(maxretr,nlayer), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate zpoint2'
        allocate( idate(maxretr), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate idate'
        allocate( itime(maxretr), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate itime'
        allocate( cak_out(maxretr,nlayer), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate cak_out'
        allocate( vapri_out(maxretr,nlayer), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate vapri_out'
        allocate( vdryair_out(maxretr,nlayer), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate vdryair_out'
        allocate( vmr_out(maxretr), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate vmr_out'
        allocate( verr_out(maxretr), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate verr_out'
        allocate( albedo_out(maxretr), stat=ierr )
        if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate albedo_out'
        cak_out(:,:)=0.
        vapri_out(:,:)=0.
        vmr_out(:)=0.
        verr_out(:)=0.
        albedo_out(:)=0.
        nretr = 0
      endif
      do nt = 1, ntime
        ndays = dble(time(nt))/3600d0/24d0
        jref = juldate(refdate, reftime)
        do nm = 1, nmeas
          do np = 1, npixel 
            ! exclude retrievals below quality cutoff
            if ( qa(np,nm,nt).lt.settings%qa_cutoff ) cycle
            ! exclude retrievals with few degrees of freedom
            if ( sum(cak(:,np,nm,nt)).lt.10. ) cycle
            write(surfclass_bin, fmt='(B0)') surfclass(np,nm,nt)
            nbits = len_trim(surfclass_bin)
            read(surfclass_bin(nbits:nbits),*) iswater
            ! if landonly exclude pixels over water
            if ( settings%landonly.and.iswater.eq.1 ) cycle
            if ( settings%lnested.eq.1 ) then
              ! if nested domain exclude pixels outside
              if ( any(lon_bnds(:,np,nm,nt).lt.nest_llx).or. &
                   any(lon_bnds(:,np,nm,nt).gt.nest_urx).or. &
                   any(lat_bnds(:,np,nm,nt).lt.nest_lly).or. &
                   any(lat_bnds(:,np,nm,nt).gt.nest_ury) ) cycle
            else
              ! if not nested exclude pixels outside global domain
              if ( any(lon_bnds(:,np,nm,nt).lt.llx).or. &
                   any(lon_bnds(:,np,nm,nt).gt.urx).or. &
                   any(lat_bnds(:,np,nm,nt).lt.lly).or. &
                   any(lat_bnds(:,np,nm,nt).gt.ury) ) cycle
            endif
            jdmeas = jref+ndays+dble(timedel(nm,nt))/1000d0/3600d0/24d0
            ! exclude retrievals if not for current day
            if ( floor(jdmeas).ne.jd ) cycle
            if ( vmr(np,nm,nt).ge.bignum ) cycle
            call caldate(jdmeas, jjjjmmdd, hhmiss)
            ! check date and time
            if (isnan(real(hhmiss)).or.(jjjjmmdd.lt.0)) then
              write(logid,*) 'WARNING: problem with itime, idate for file, nmeas, npixel, jdmeas = ',&
                               trim(filereadlist(n)), nm, np, jdmeas
              cycle
            endif
            nretr = nretr + 1
            idate(nretr) = jjjjmmdd
            itime(nretr) = hhmiss
            ! corners ordered anticlockwise from lower left
            xpoints(nretr,:) = lon_bnds(:,np,nm,nt)
            ypoints(nretr,:) = lat_bnds(:,np,nm,nt)
            ! pressure levels are equipressure apart
            zpoint1(nretr,1) = psurf(np,nm,nt)
            zpoint2(nretr,1) = psurf(np,nm,nt) - pdel(np,nm,nt)
            do nl = 2, nlayer
              zpoint1(nretr,nl) = zpoint1(nretr,nl-1) - pdel(np,nm,nt)
              zpoint2(nretr,nl) = zpoint2(nretr,nl-1) - pdel(np,nm,nt)
            end do
            cak_out(nretr,:) = cak(:,np,nm,nt)
            albedo_out(nretr) = albedo(np,nm,nt)
            vapri_out(nretr,:) = vapri(:,np,nm,nt)
            vdryair_out(nretr,:) = vdryair(:,np,nm,nt)
            vmr_out(nretr) = vmr(np,nm,nt)
            ! uncertainties for XCH4 based on single sounding precision due to measurement noise -> 
            ! for overall uncertainty estimate multiply by factor 2 to reflect scatter of single sounding errors
            verr_out(nretr) = 2.*verr(np,nm,nt)
          end do ! npixel
        end do ! nmeas
      end do ! ntime

      write(logid,*) 'number of retrievals: ',nretr

      if ( allocated(lat_bnds) )      deallocate(lat_bnds)
      if ( allocated(lon_bnds) )      deallocate(lon_bnds)
      if ( allocated(psurf) )         deallocate(psurf)
      if ( allocated(pdel) )          deallocate(pdel)
      if ( allocated(vapri) )         deallocate(vapri)
      if ( allocated(vdryair) )       deallocate(vdryair)
      if ( allocated(cak) )           deallocate(cak)
      if ( allocated(albedo) )        deallocate(albedo)
      if ( allocated(timedel) )       deallocate(timedel)
      if ( allocated(time) )          deallocate(time)
      if ( allocated(qa) )            deallocate(qa)
      if ( allocated(vmr) )           deallocate(vmr)
      if ( allocated(verr) )          deallocate(verr)
      if ( allocated(surfclass) )     deallocate(surfclass)

    end do  ! nread

    ! reverse direction of vertical coordinate
    ! raw data ordered from top of atmosphere to surface
    ! which is opposite to the pressure levels calculated (zpoint)
    allocate( cak_tmp(maxretr,nlayer), stat=ierr )
    if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate cak_tmp'
    allocate( vapri_tmp(maxretr,nlayer), stat=ierr )
    if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate vapri_tmp'
    allocate( vdryair_tmp(maxretr,nlayer), stat=ierr )
    if ( ierr.ne.0 ) write(logid,*) 'ERROR: cannot allocate vdryair_tmp'

    do nl = 1, nlayer
      cak_tmp(:,nl) = cak_out(:,(nlayer-nl+1))
      vapri_tmp(:,nl) = vapri_out(:,(nlayer-nl+1))
      vdryair_tmp(:,nl) = vdryair_out(:,(nlayer-nl+1))
    end do
    cak_out = cak_tmp
    vapri_out = vapri_tmp
    vdryair_out = vdryair_tmp
    deallocate( cak_tmp, vapri_tmp, vdryair_tmp)

    ! calculate column prior (units mol/m2)
    allocate( cpri(maxretr), stat=ierr )
    cpri(1:nretr) = sum(vapri_out(1:nretr,:), dim=2)

    ! calculate column prior convolved with averaging kernel (units mol/m2) 
    allocate( cakpri(maxretr), stat=ierr )
    do nm = 1, nretr
      cakpri(nm) = dot_product(vapri_out(nm,:), cak_out(nm,:))
    end do

    ! apply albedo correction used in Official retrievals to cak, cpri and cakpri 
    ! this is polynomial function order of 11 (Soumyajit Mandal, SRON)
!    const =  1.0282413517102085
!    poly(1) = -0.4717177565799524
!    poly(2) = 2.1491179428159324
!    poly(3) = 16.952582059603163
!    poly(4) = -250.66075995913351
!    poly(5) = 1307.5191440027886
!    poly(6) =-3791.6452849323673
!    poly(7) =6783.218182991253
!    poly(8) =-7689.387517819326
!    poly(9) = 5404.118866385074
!    poly(10) =-2154.928509865762
!    poly(11) =373.4579485937264
!    if ( settings%albedo_corr ) then
!      open(100,file=trim(settings%path_obsout)//'aldbedo_corr_'//adate//'.txt', &
!            status='replace',action='write')
!      do nm = 1, nretr
!        alb = albedo_out(nm)
!        factor_corr = const + poly(1)*alb + poly(2)*alb**2 + poly(3)*alb**3 + poly(4)*alb**4 + &
!                       poly(5)*alb**5 + poly(6)*alb**6 + poly(7)*alb**7 + poly(8)*alb**8 + &
!                       poly(9)*alb**9 + poly(10)*alb**10 + poly(11)*alb**11
!        cakpri(nm) = factor_corr*cakpri(nm)
!        cpri(nm) = factor_corr*cpri(nm)
!        cak_out(nm,:) = factor_corr*cak_out(nm,:)
!        write(100,'(F6.4,1X,F6.4)') factor_corr, alb
!      end do
!      close(100)
!    endif
    
    ! calculate column dry air (units mol/m2)
    allocate( cdryair(maxretr), stat=ierr )
    cdryair(1:nretr) = sum(vdryair_out(1:nretr,:), dim=2)

    ! average pixels
    ! --------------

    if ( settings%avg_pixels ) then
      ! nretr is number of individual observations
      ! after averaging numbox is number of super observations
      call average(settings, idate, itime, xpoints, ypoints, zpoint1, zpoint2, vmr_out, verr_out, &
                   vapri_out, cpri, cakpri, cak_out, cdryair, vdryair_out, albedo_out, nretr, &
                   ncoord, nlayer, numbox)
    else
      ! no averaging
      numbox = nretr
    endif

    ! write output data
    ! -----------------

    ! adjust itime 
    ! needed so that with flexpart rounding to lsynctime all retrievals are still for current day
    do nm = 1, numbox
      if ( itime(nm).gt.(235959-lsynctime*100/60) ) then
        itime(nm) = 235950-lsynctime*100/60
      endif
    end do

    ! write command file
    call prep_command(settings,jd)

    ! write outgrid file
    call prep_outgrid(settings,jd)

    ! write ageclasses file
    call prep_ageclass(settings,jd)

    ! write pathnames file
    call prep_pathnames(settings,jd)

    ! units of output data: dryair, cdryair, cpri, vmr
    units(1) = 'mol/m2'
    units(2) = 'mol/m2'
    units(3) = 'mol/m2'
    units(4) = 'ppbv'

    ! write releases data for flexpart
    call write_releases(settings,numbox,nlayer,ncoord,xpoints,ypoints,zpoint1,zpoint2,&
                        idate,itime,cak_out,vdryair_out,specnum,units)

    ! write retrieval data for flexinvert
    call write_retrieval(settings,numbox,ncoord,xpoints,ypoints,idate,itime,&
                         cpri,cakpri,vmr_out,verr_out,cdryair,albedo_out,units)

    ! copy standard input files to options folder
    write(aspec,fmt='(I3.3)') specnum
    call system('mkdir -p '//trim(settings%path_options)//adate//'/options/SPECIES/')
    filename = trim(settings%path_flexpart)//'options/SPECIES/SPECIES_'//aspec
    call system('cp '//trim(filename)//' '//trim(settings%path_options)//adate//'/options/SPECIES/')
    call system('cp '//trim(settings%path_flexpart)//'options/*.dat'//' '//trim(settings%path_options)//adate//'/options/')
    call system('cp '//trim(settings%path_flexpart)//'options/*.t'//' '//trim(settings%path_options)//adate//'/options/')

    ! prepare for next day
    ! -------------------- 

    if ( allocated(xpoints) )     deallocate(xpoints)
    if ( allocated(ypoints) )     deallocate(ypoints)
    if ( allocated(zpoint1) )     deallocate(zpoint1)
    if ( allocated(zpoint2) )     deallocate(zpoint2)
    if ( allocated(idate) )       deallocate(idate)
    if ( allocated(itime) )       deallocate(itime)
    if ( allocated(cak_out) )     deallocate(cak_out)
    if ( allocated(vapri_out) )   deallocate(vapri_out)
    if ( allocated(vdryair_out) ) deallocate(vdryair_out)
    if ( allocated(vmr_out) )     deallocate(vmr_out)
    if ( allocated(albedo_out) )  deallocate(albedo_out)
    if ( allocated(verr_out) )    deallocate(verr_out)
    if ( allocated(cpri) )        deallocate(cpri)
    if ( allocated(cakpri) )      deallocate(cakpri)
    if ( allocated(cdryair) )     deallocate(cdryair)

    ! iterate over days
    jd = jd + 1d0

  end do read_loop

  if ( allocated(filelist) )     deallocate(filelist)
  if ( allocated(filereadlist) ) deallocate(filereadlist)


end subroutine get_tropomi

